"""Unit tests."""

import json
import os
import shutil
from unittest.mock import patch

import numpy as np
import pandas as pd
import pytest
import statsmodels.api as sm

from acro import ACRO, acro_tables, add_constant, add_to_acro, record, utils
from acro.acro_tables import rounded_survival_table
from acro.record import Records, load_records

# pylint: disable=redefined-outer-name,too-many-lines

PATH: str = "RES_PYTEST"


@pytest.fixture
def data() -> pd.DataFrame:
    """Load test data."""
    path = os.path.join("data", "test_data.dta")
    data = pd.read_stata(path)
    return data


@pytest.fixture
def acro() -> ACRO:
    """Initialise ACRO."""
    return ACRO(suppress=True)


def test_add_backticks():
    """Test the add_backticks helper function."""
    # Test simple string without spaces (no backticks added)
    assert acro_tables.add_backticks("foo") == "foo"

    # Test string with spaces (backticks should be added)
    assert acro_tables.add_backticks("foo bar") == "`foo bar`"

    # Test string already with backticks (no change)
    assert acro_tables.add_backticks("`foo bar`") == "`foo bar`"

    # Test multiple spaces
    assert acro_tables.add_backticks("foo bar baz") == "`foo bar baz`"


def test_crosstab_with_spaces_in_variable_names(data, acro):
    """Test crosstab with spaces in column names (Issue #305)."""
    # Create a test dataframe with a column name containing spaces
    test_data = data.copy()
    test_data["grant type with spaces"] = test_data["grant_type"]
    test_data["year of study"] = test_data["year"]

    # This should handle spaces in variable names correctly
    # first check is that it behaves the same as pandas without suppression
    acro.suppress = False
    pandas_nospace_version = pd.crosstab(data["year"], data["grant_type"], margins=True)
    acro_with_spaces_version = acro.crosstab(
        test_data["year of study"], test_data["grant type with spaces"], margins=True
    )
    assert (
        acro_with_spaces_version.to_numpy() == pandas_nospace_version.to_numpy()
    ).all()
    # Verify that suppression was not applied in this case
    assert acro.results.get_index(-1).status == "fail"

    # Verify the crosstab was created successfully

    # turn suppression back on, run rest of checks
    acro.suppress = True
    result = acro.crosstab(
        test_data["year of study"], test_data["grant type with spaces"], margins=True
    )
    assert isinstance(result, pd.DataFrame)
    assert not result.empty

    # Verify that suppression was applied in second case
    assert acro.results.get_index(-1).status == "review"


def test_crosstab_without_suppression(data):
    """Crosstab threshold without automatic suppression."""
    acro = ACRO(suppress=False)
    _ = acro.crosstab(data.year, data.grant_type)
    output = acro.results.get_index(0)
    correct_summary: str = "fail; threshold: 6 cells may need suppressing; "
    assert output.summary == correct_summary
    assert 48 == output.output[0]["R/G"].sum()


def test_crosstab_with_aggfunc_mode(data):
    """Crosstab threshold without automatic suppression."""
    acro = ACRO(suppress=False)
    _ = acro.crosstab(
        data.year, data.grant_type, values=data.inc_grants, aggfunc="mode"
    )
    output = acro.results.get_index(0)
    correct_summary: str = "fail; all-values-are-same: 1 cells may need suppressing; "
    assert output.summary == correct_summary
    assert 913000 == output.output[0]["R/G"].iat[0]


def test_crosstab_with_aggfunc_sum(data, acro):
    """Test the crosstab with two columns and aggfunc sum."""
    acro = ACRO(suppress=False)
    _ = acro.crosstab(
        data.year,
        [data.grant_type, data.survivor],
        values=data.inc_grants,
        aggfunc="sum",
    )
    _ = acro.crosstab(
        [data.grant_type, data.survivor],
        data.year,
        values=data.inc_grants,
        aggfunc="sum",
    )
    acro.add_exception("output_0", "Let me have it")
    acro.add_exception("output_1", "I need this output")
    results: Records = acro.finalise(PATH)
    output_0 = results.get_index(0)
    output_1 = results.get_index(1)
    comment_0 = (
        "Empty columns: ('N', 'Dead in 2015'), ('R/G', 'Dead in 2015') were deleted."
    )
    comment_1 = (
        "Empty rows: ('N', 'Dead in 2015'), ('R/G', 'Dead in 2015') were deleted."
    )
    assert output_0.comments == [comment_0]
    assert output_1.comments == [comment_1]
    shutil.rmtree(PATH)


def test_crosstab_threshold(data, acro):
    """Crosstab threshold test."""
    _ = acro.crosstab(data.year, data.grant_type)
    output = acro.results.get_index(0)
    total_nan: int = output.output[0]["R/G"].isnull().sum()
    assert total_nan == 6
    positions = output.sdc["cells"]["threshold"]
    for pos in positions:
        row, col = pos
        assert np.isnan(output.output[0].iloc[row, col])
    acro.add_exception("output_0", "Let me have it")
    results: Records = acro.finalise(PATH)
    correct_summary: str = "review; threshold: 6 cells suppressed; "
    output = results.get_index(0)
    assert output.summary == correct_summary
    shutil.rmtree(PATH)


def test_crosstab_multiple(data, acro):
    """Crosstab multiple rule test."""
    _ = acro.crosstab(
        data.year, data.grant_type, values=data.inc_grants, aggfunc="mean"
    )
    acro.add_exception("output_0", "Let me have it")
    results: Records = acro.finalise(PATH)
    correct_summary: str = (
        "review; threshold: 7 cells suppressed; p-ratio: 2 cells suppressed; "
        "nk-rule: 1 cells suppressed; "
    )
    output = results.get_index(0)
    assert output.summary == correct_summary
    shutil.rmtree(PATH)


def test_negatives(data, acro):
    """Pivot table and Crosstab with negative values."""
    data.loc[0:10, "inc_grants"] = -10
    _ = acro.crosstab(
        data.year, data.grant_type, values=data.inc_grants, aggfunc="mean"
    )
    _ = acro.pivot_table(
        data, index=["grant_type"], values=["inc_grants"], aggfunc=["mean", "std"]
    )
    acro.add_exception("output_0", "Let me have it")
    acro.add_exception("output_1", "I want this")
    results: Records = acro.finalise(PATH)
    correct_summary: str = "review; negative values found"
    output_0 = results.get_index(0)
    output_1 = results.get_index(1)
    assert output_0.summary == correct_summary
    assert output_1.summary == correct_summary
    shutil.rmtree(PATH)


def test_pivot_table_without_suppression(data):
    """Pivot table without automatic suppression."""
    acro = ACRO(suppress=False)
    _ = acro.pivot_table(
        data, index=["grant_type"], values=["inc_grants"], aggfunc=["mean", "std"]
    )
    output_0 = acro.results.get_index(0)
    assert 36293992.0 == output_0.output[0]["mean"]["inc_grants"].sum()
    assert "pass" == output_0.summary


def test_pivot_table_pass(data, acro):
    """Pivot table pass test."""
    _ = acro.pivot_table(
        data, index=["grant_type"], values=["inc_grants"], aggfunc=["mean", "std"]
    )
    results: Records = acro.finalise(PATH)
    correct_summary: str = "pass"
    output_0 = results.get_index(0)
    assert output_0.summary == correct_summary
    shutil.rmtree(PATH)


def test_pivot_table_cols(data, acro):
    """Pivot table with columns test."""
    _ = acro.pivot_table(
        data,
        index=["grant_type"],
        columns=["year"],
        values=["inc_grants"],
        aggfunc=["mean", "std"],
    )
    acro.add_exception("output_0", "Let me have it")
    results: Records = acro.finalise(PATH)
    correct_summary: str = (
        "review; threshold: 14 cells suppressed; "
        "p-ratio: 4 cells suppressed; nk-rule: 2 cells suppressed; "
    )
    output_0 = results.get_index(0)
    assert output_0.summary == correct_summary
    shutil.rmtree(PATH)


def test_pivot_table_with_aggfunc_sum(data, acro):
    """Test the pivot table with two columns and aggfunc sum."""
    acro = ACRO(suppress=False)
    _ = acro.pivot_table(
        data,
        index="year",
        columns=["grant_type", "survivor"],
        values="inc_grants",
        aggfunc="sum",
    )
    _ = acro.pivot_table(
        data,
        index=["grant_type", "survivor"],
        columns="year",
        values="inc_grants",
        aggfunc="sum",
    )
    acro.add_exception("output_0", "Let me have it")
    acro.add_exception("output_1", "I need this output")
    results: Records = acro.finalise(PATH)
    output_0 = results.get_index(0)
    output_1 = results.get_index(1)
    comment_0 = (
        "Empty columns: ('N', 'Dead in 2015'), ('R/G', 'Dead in 2015') were deleted."
    )
    comment_1 = (
        "Empty rows: ('N', 'Dead in 2015'), ('R/G', 'Dead in 2015') were deleted."
    )
    assert output_0.comments == [comment_0]
    assert output_1.comments == [comment_1]
    shutil.rmtree(PATH)


def test_ols(data, acro):
    """Ordinary Least Squares test."""
    new_df = data[["inc_activity", "inc_grants", "inc_donations", "total_costs"]]
    new_df = new_df.dropna()
    # OLS too few Dof
    endog = new_df.inc_activity.iloc[0:10]
    exog = new_df[["inc_grants", "inc_donations", "total_costs"]].iloc[0:10]
    exog = add_constant(exog)
    results = acro.ols(endog, exog)
    assert results.df_resid == 6
    res = acro.results.get_index(-1)
    summary = res.summary.split(";")
    assert summary[0] == "fail"
    acro.remove_output(res.uid)

    # OLS
    endog = new_df.inc_activity
    exog = new_df[["inc_grants", "inc_donations", "total_costs"]]
    exog = add_constant(exog)
    results = acro.ols(endog, exog)
    assert results.df_resid == 807
    assert results.rsquared == pytest.approx(0.894, 0.001)
    # OLSR
    results = acro.olsr(
        formula="inc_activity ~ inc_grants + inc_donations + total_costs", data=new_df
    )
    assert results.df_resid == 807
    assert results.rsquared == pytest.approx(0.894, 0.001)
    # Finalise
    results = acro.finalise(PATH)
    correct_summary: str = "pass; dof=807.0 >= 10"
    output_0 = results.get_index(0)
    output_1 = results.get_index(1)
    assert output_0.summary == correct_summary
    assert output_1.summary == correct_summary
    shutil.rmtree(PATH)


def test_probit_logit(data, acro):
    """Probit and Logit tests."""
    new_df = data[
        ["survivor", "inc_activity", "inc_grants", "inc_donations", "total_costs"]
    ]
    new_df = new_df.dropna()
    endog = new_df["survivor"].astype("category").cat.codes  # numeric
    endog.name = "survivor"
    exog = new_df[["inc_activity", "inc_grants", "inc_donations", "total_costs"]]
    exog = add_constant(exog)
    # Probit
    results = acro.probit(endog, exog)
    assert results.df_resid == 806
    assert results.prsquared == pytest.approx(0.208, 0.01)
    # Logit
    results = acro.logit(endog, exog)
    assert results.df_resid == 806
    assert results.prsquared == pytest.approx(0.214, 0.01)
    # ProbitR
    new_df["survivor"] = new_df["survivor"].astype("category").cat.codes
    results = acro.probitr(
        formula="survivor ~ inc_activity + inc_grants + inc_donations + total_costs",
        data=new_df,
    )
    assert results.df_resid == 806
    assert results.prsquared == pytest.approx(0.208, 0.01)
    # LogitR
    results = acro.logitr(
        formula="survivor ~ inc_activity + inc_grants + inc_donations + total_costs",
        data=new_df,
    )
    assert results.df_resid == 806
    assert results.prsquared == pytest.approx(0.214, 0.01)
    # Finalise
    results = acro.finalise(PATH)
    correct_summary: str = "pass; dof=806.0 >= 10"
    output_0 = results.get_index(0)
    output_1 = results.get_index(1)
    output_2 = results.get_index(2)
    output_3 = results.get_index(3)
    assert output_0.summary == correct_summary
    assert output_1.summary == correct_summary
    assert output_2.summary == correct_summary
    assert output_3.summary == correct_summary
    shutil.rmtree(PATH)


def test_finalise_excel(data, acro):
    """Finalise excel test."""
    _ = acro.crosstab(data.year, data.grant_type)
    acro.add_exception("output_0", "Let me have it")
    with open("foo.txt", "w") as file:
        file.write("Your text goes here")
    acro.custom_output("foo.txt")

    results: Records = acro.finalise(PATH, "xlsx")
    output_0 = results.get_index(0)
    filename = os.path.normpath(f"{PATH}/results.xlsx")
    load_data = pd.read_excel(filename, sheet_name=output_0.uid)
    correct_cell: str = "_ = acro.crosstab(data.year, data.grant_type)"
    assert load_data.iloc[0, 0] == "Command"
    assert load_data.iloc[0, 1] == correct_cell
    shutil.rmtree(PATH)


def test_output_removal(data, acro, monkeypatch):
    """Output removal and print test."""
    _ = acro.crosstab(data.year, data.grant_type)
    _ = acro.crosstab(data.year, data.grant_type)
    _ = acro.crosstab(data.year, data.grant_type)
    exceptions = ["I want it", "Let me have it", "Please!"]
    monkeypatch.setattr("builtins.input", lambda _: exceptions.pop(0))
    results: Records = acro.finalise(PATH)
    output_0 = results.get("output_0")
    output_1 = results.get("output_1")
    shutil.rmtree(PATH)
    # remove something that is there
    acro.remove_output(output_0.uid)
    results = acro.finalise(PATH)
    correct_summary: str = "review; threshold: 6 cells suppressed; "
    keys = results.get_keys()
    assert output_0.uid not in keys
    assert output_1.uid in keys
    assert output_1.summary == correct_summary
    acro.print_outputs()
    # remove something that is not there
    with pytest.raises(ValueError, match="unable to remove 123, key not found"):
        acro.remove_output("123")
    shutil.rmtree(PATH)


def test_load_output():
    """Empty or invalid array when loading output."""
    with pytest.raises(ValueError, match="error loading output"):
        record.load_output(PATH, [])
    val = record.load_output(PATH, ["nosuchfile.xxx"])
    assert val == ["nosuchfile.xxx"]


def test_finalise_invalid(data, acro):
    """Invalid output format when finalising."""
    _ = acro.crosstab(data.year, data.grant_type)
    output_0 = acro.results.get_index(0)
    output_0.exception = "Let me have it"
    with pytest.raises(ValueError, match="Invalid file extension.*"):
        _ = acro.finalise(PATH, "123")


def test_finalise_json(data, acro):
    """Finalise json test."""
    _ = acro.crosstab(data.year, data.grant_type)
    acro.add_exception("output_0", "Let me have it")
    # write JSON
    result: Records = acro.finalise(PATH, "json")
    # load JSON
    loaded: Records = load_records(PATH)
    orig = result.get_index(0)
    read = loaded.get_index(0)
    print("*****************************")
    print(orig)
    print("*****************************")
    print(read)
    print("*****************************")
    # check equal
    assert orig.uid == read.uid
    assert orig.status == read.status
    assert orig.output_type == read.output_type
    assert orig.properties == read.properties
    assert orig.sdc == read.sdc
    assert orig.command == read.command
    assert orig.summary == read.summary
    assert orig.comments == read.comments
    assert orig.timestamp == read.timestamp
    # check SDC outcome DataFrame
    orig_df = orig.output[0].reset_index()
    read_df = read.output[0]
    pd.testing.assert_frame_equal(
        orig_df, read_df, check_names=False, check_dtype=False
    )
    # test reading JSON
    with open(os.path.normpath(f"{PATH}/results.json"), encoding="utf-8") as file:
        json_data = json.load(file)
    results: dict = json_data["results"]
    assert results[orig.uid]["files"][0]["name"] == f"{orig.uid}_0.csv"
    shutil.rmtree(PATH)


def test_rename_output(data, acro):
    """Output renaming test."""
    _ = acro.crosstab(data.year, data.grant_type)
    _ = acro.crosstab(data.year, data.grant_type)
    acro.add_exception("output_0", "Let me have it")
    acro.add_exception("output_1", "I want this")
    results: Records = acro.finalise(PATH)
    output_0 = results.get_index(0)
    orig_name = output_0.uid
    new_name = "cross_table"
    acro.rename_output(orig_name, new_name)
    shutil.rmtree(PATH)
    results = acro.finalise(PATH)
    assert output_0.uid == new_name
    assert orig_name not in results.get_keys()
    assert os.path.exists(f"{PATH}/{new_name}_0.csv")
    # rename an output that doesn't exist
    with pytest.raises(ValueError, match="unable to rename 123, key not found"):
        acro.rename_output("123", "name")
    # rename an output to another that already exists
    with pytest.raises(ValueError, match="unable to rename, cross_table .* exists"):
        acro.rename_output("output_1", "cross_table")
    shutil.rmtree(PATH)


def test_add_comments(data, acro):
    """Adding comments to output test."""
    _ = acro.crosstab(data.year, data.grant_type)
    acro.add_exception("output_0", "Let me have it")
    results: Records = acro.finalise(PATH)
    output_0 = results.get_index(0)
    assert output_0.comments == []
    comment = "This is a cross table between year and grant_type"
    acro.add_comments(output_0.uid, comment)
    assert output_0.comments == [comment]
    comment_1 = "6 cells were suppressed"
    acro.add_comments(output_0.uid, comment_1)
    assert output_0.comments == [comment, comment_1]
    # add a comment to something that is not there
    with pytest.raises(ValueError, match="unable to find 123, key not found"):
        acro.add_comments("123", "comment")
    shutil.rmtree(PATH)


def test_custom_output(acro):
    """Adding an unsupported output to the results dictionary test."""
    filename = "notebooks/XandY.jpeg"
    file_path = os.path.normpath(filename)
    acro.custom_output(filename)
    acro.add_exception("output_0", "Let me have it")
    results: Records = acro.finalise(path=PATH)
    output_0 = results.get_index(0)
    assert output_0.output == [file_path]
    assert os.path.exists(os.path.normpath(f"{PATH}/XandY.jpeg"))
    shutil.rmtree(PATH)


def test_missing(data, acro, monkeypatch):
    """Pivot table and Crosstab with negative values."""
    acro_tables.CHECK_MISSING_VALUES = True
    acro.suppress = False
    data.loc[0:10, "inc_grants"] = np.nan
    _ = acro.crosstab(
        data.year, data.grant_type, values=data.inc_grants, aggfunc="mean"
    )
    _ = acro.pivot_table(
        data, index=["grant_type"], values=["inc_grants"], aggfunc=["mean", "std"]
    )
    exceptions = ["I want it", "Let me have it"]
    monkeypatch.setattr("builtins.input", lambda _: exceptions.pop(0))
    results: Records = acro.finalise(PATH, interactive=True)
    correct_summary: str = "review; missing values found"
    output_0 = results.get_index(0)
    output_1 = results.get_index(1)
    assert output_0.summary == correct_summary
    assert output_1.summary == correct_summary
    assert output_0.exception == "I want it"
    assert output_1.exception == "Let me have it"
    shutil.rmtree(PATH)


def test_suppression_error(caplog):
    """Apply suppression type error test."""
    table_data = {"col1": [1, 2], "col2": [3, 4]}
    mask_data = {"col1": [np.nan, True], "col2": [True, True]}
    table = pd.DataFrame(data=table_data)
    masks = {"test": pd.DataFrame(data=mask_data)}
    acro_tables.apply_suppression(table, masks)
    assert "problem mask test is not binary" in caplog.text


def test_adding_exception(acro):
    """Adding an exception to an output that doesn't exist test."""
    with pytest.raises(ValueError, match="unable to add exception: output_0 .*"):
        acro.add_exception("output_0", "Let me have it")


def test_add_to_acro(data, monkeypatch):
    """Add an output generated without acro to an acro object and create results file."""
    # create a cross tabulation using pandas
    table = pd.crosstab(data.year, data.grant_type)
    # save the output to a file and add this file to a directory
    src_path = "test_add_to_acro"
    dest_path = "sdc_results"
    file_path = "crosstab.pkl"
    if not os.path.exists(src_path):  # pragma no cover
        table.to_pickle(file_path)
        os.mkdir(src_path)
        shutil.move(file_path, src_path, copy_function=shutil.copytree)
    # add exception to the output
    exception = ["I want it"]
    monkeypatch.setattr("builtins.input", lambda _: exception.pop(0))
    # add the output to acro
    add_to_acro(src_path, dest_path)
    assert "results.json" in os.listdir(dest_path)
    assert "crosstab.pkl" in os.listdir(dest_path)


def test_prettify_tablestring(data):
    """Test prettifying string version of table."""
    mydata = data
    # take subsets for brevity
    mydata = mydata[(mydata["charity"].str[0] == "W")]
    mydata = mydata[mydata["year"] < 2012]
    correct = (
        "----------------------------------------------------------------------|\n"
        "grant_type                               |N          |R    |R/G       |\n"
        "status                                   |successful |dead |successful|\n"
        "year charity                             |           |     |          |\n"
        "----------------------------------------------------------------------|\n"
        "2010 WWF                                 | 0         | 0   | 1        |\n"
        "     Walsall domestic violence forum ltd | 0         | 1   | 0        |\n"
        "     Will Woodlands                      | 1         | 0   | 0        |\n"
        "     Worcestershire Lifestyles (Dead)    | 0         | 1   | 0        |\n"
        "2011 WWF                                 | 0         | 0   | 1        |\n"
        "     Walsall domestic violence forum ltd | 0         | 1   | 0        |\n"
        "     Will Woodlands                      | 1         | 0   | 0        |\n"
        "     Worcestershire Lifestyles (Dead)    | 0         | 1   | 0        |\n"
        "----------------------------------------------------------------------|\n"
    )
    complex_str = utils.prettify_table_string(
        pd.crosstab([mydata.year, mydata.charity], [mydata.grant_type, mydata.status])
    )
    assert complex_str == correct, f"got:\n{complex_str}\nexpected:\n{correct}\n"

    correct2 = (
        "------------------------|\n"
        "grant_type  |N  |R  |R/G|\n"
        "year        |   |   |   |\n"
        "------------------------|\n"
        "2010        |1  |2  |1  |\n"
        "2011        |1  |2  |1  |\n"
        "------------------------|\n"
    )
    simple_str = utils.prettify_table_string(
        pd.crosstab([mydata.year], [mydata.grant_type])
    )
    assert simple_str == correct2, f"got:\n{simple_str}\nexpected:\n{correct2}\n"

    # test spaces in variable names dealt with
    correct3 = (
        "---------------------------------------|\n"
        "survivor  |Dead_in_2015  |Alive_in_2015|\n"
        "year      |              |             |\n"
        "---------------------------------------|\n"
        "2010      |2             |2            |\n"
        "2011      |2             |2            |\n"
        "---------------------------------------|\n"
    )
    nospaces__str = utils.prettify_table_string(
        pd.crosstab([mydata.year], [mydata.survivor])
    )
    assert nospaces__str == correct3, f"got:\n{nospaces__str}\nexpected:\n{correct3}\n"


def test_hierachical_aggregation(data, acro):
    """Should work with hierarchies in rows/columns."""
    acro.suppress = False
    the_data = data[data.grant_type != "G"]
    result = acro.crosstab(
        [the_data.year, the_data.survivor],
        [the_data.grant_type],
        values=the_data.inc_activity,
        aggfunc="sum",
    )
    res = utils.prettify_table_string(result)
    correct = (
        "------------------------------------------------------------|\n"
        "grant_type          |N           |R             |R/G        |\n"
        "year survivor       |            |              |           |\n"
        "------------------------------------------------------------|\n"
        "2010 Dead in 2015   |       0.0  |1.723599e+07  |        0.0|\n"
        "     Alive in 2015  |52865600.0  |6.791129e+08  | 24592000.0|\n"
        "2011 Dead in 2015   |       0.0  |1.890400e+07  |        0.0|\n"
        "     Alive in 2015  |66714452.0  |1.002141e+09  | 86171000.0|\n"
        "2012 Dead in 2015   |       0.0  |2.616444e+07  |        0.0|\n"
        "     Alive in 2015  |64777124.0  |1.013167e+09  |107716000.0|\n"
        "2013 Dead in 2015   |       0.0  |2.913558e+07  |        0.0|\n"
        "     Alive in 2015  |86806336.0  |1.048305e+09  |104197000.0|\n"
        "2014 Dead in 2015   |       0.0  |3.074519e+07  |        0.0|\n"
        "     Alive in 2015  |74486664.0  |1.035069e+09  |106287000.0|\n"
        "2015 Dead in 2015   |       0.0  |1.488808e+07  |        0.0|\n"
        "     Alive in 2015  |56155352.0  |9.932494e+08  |105224000.0|\n"
        "------------------------------------------------------------|\n"
    )
    assert res.split() == correct.split(), f"got\n{res}\nexpected\n{correct}\n"


def test_single_values_column(data, acro):
    """Pandas does not allows multiple arrays for values."""
    with pytest.raises(ValueError, match=".*specify a single values column.*"):
        _ = acro.crosstab(
            data.year,
            data.grant_type,
            values=[data.inc_activity, data.inc_activity],
            aggfunc="mean",
        )
    with pytest.raises(ValueError, match=".*specify a single values column.*"):
        _ = acro.crosstab(data.year, data.grant_type, values=None, aggfunc="mean")


def test_surv_func(acro):
    """Test survival tables and plots."""
    # Load real data but with fallback to mock if network fails
    try:
        data = sm.datasets.get_rdataset("flchain", "survival").data
    except Exception:
        # Fallback to mock data if network is unavailable
        np.random.seed(42)
        mock_data = pd.DataFrame(
            {
                "futime": np.random.exponential(100, 500),
                "death": np.random.binomial(1, 0.3, 500),
                "sex": np.random.choice(["F", "M"], 500),
            }
        )
        data = mock_data
        # Skip the exact assertion when using mock data
        skip_exact_assertion = True
    else:
        skip_exact_assertion = False

    data = data.loc[data.sex == "F", :]
    # table
    _ = acro.surv_func(data.futime, data.death, output="table")
    output = acro.results.get_index(0)
    correct_summary: str = "review; threshold: 3864 cells suppressed; "
    assert output.summary == correct_summary

    if not skip_exact_assertion:
        assert output.summary == correct_summary
    else:
        # Just verify the output contains "fail" and "cells suppressed"
        assert "fail" in output.summary
        assert "cells suppressed" in output.summary

    # plot
    filename = os.path.normpath("acro_artifacts/kaplan-meier_0.png")
    _ = acro.surv_func(data.futime, data.death, output="plot")
    assert os.path.exists(filename)
    acro.add_exception("output_0", "I need this")
    acro.add_exception("output_1", "Let me have it")

    # neither table nor plot
    foo = acro.surv_func(data.futime, data.death, output="something_else")
    assert foo is None

    results: Records = acro.finalise(path=PATH)
    output_1 = results.get_index(1)
    assert output_1.output == [filename]
    shutil.rmtree(PATH)


def test_rounded_survival_table():
    """Test the rounded_survival_table function for survival analysis."""
    # Create a minimal survival table with required columns
    survival_table = pd.DataFrame(
        {
            "Surv prob": [1.0, 0.95, 0.90, 0.85, 0.80],
            "num at risk": [100, 95, 85, 75, 60],
            "num events": [0, 5, 10, 10, 15],
        }
    )

    # Apply rounded_survival_table
    result = rounded_survival_table(survival_table.copy())

    # Check that it has the rounded_survival_fun column
    assert "rounded_survival_fun" in result.columns
    assert len(result) == 5

    # Check that values are reasonable (between 0 and 1)
    assert all(
        (result["rounded_survival_fun"] >= 0) & (result["rounded_survival_fun"] <= 1)
    )


def test_zeros_are_not_disclosive(data, acro):
    """Test that zeros are handled as not disclosive when `zeros_are_disclosive=False`."""
    acro_tables.ZEROS_ARE_DISCLOSIVE = False
    _ = acro.pivot_table(
        data,
        index=["grant_type"],
        columns=["year"],
        values=["inc_grants"],
        aggfunc=["mean", "std"],
    )
    acro.add_exception("output_0", "Let me have it")
    results: Records = acro.finalise(PATH)
    correct_summary: str = (
        "review; threshold: 14 cells suppressed; "
        "p-ratio: 2 cells suppressed; nk-rule: 2 cells suppressed; "
    )
    output_0 = results.get_index(0)
    assert output_0.summary == correct_summary
    shutil.rmtree(PATH)


def test_crosstab_with_totals_without_suppression(data, acro):
    """Test the crosstab with margins is true and suppression is false."""
    acro.suppress = False
    _ = acro.crosstab(data.year, data.grant_type, margins=True)
    output = acro.results.get_index(0)
    assert 153 == output.output[0]["All"].iat[0]

    total_rows = output.output[0].iloc[-1, 0:4].sum()
    total_cols = output.output[0].loc[2010:2015, "All"].sum()
    assert 918 == total_rows == total_cols == output.output[0]["All"].iat[6]


def test_crosstab_with_totals_with_suppression(data, acro):
    """Test the crosstab with both margins and suppression are true."""
    _ = acro.crosstab(data.year, data.grant_type, margins=True)
    output = acro.results.get_index(0)
    assert 145 == output.output[0]["All"].iat[0]

    total_rows = output.output[0].iloc[-1, 0:3].sum()
    total_cols = output.output[0].loc[2010:2015, "All"].sum()
    assert 870 == total_cols == total_rows == output.output[0]["All"].iat[6]
    assert "R/G" not in output.output[0].columns


def test_crosstab_with_totals_with_suppression_hierarchical(data, acro):
    """Test the crosstab with both margins and suppression are true."""
    _ = acro.crosstab(
        [data.year, data.survivor], [data.grant_type, data.status], margins=True
    )
    output = acro.results.get_index(0)
    assert 47 == output.output[0]["All"].iat[0]

    total_rows = (output.output[0].loc[("All", ""), :].sum()) - output.output[0][
        "All"
    ].iat[12]
    total_cols = (output.output[0].loc[:, "All"].sum()) - output.output[0]["All"].iat[
        12
    ]
    assert total_cols == total_rows == output.output[0]["All"].iat[12] == 852
    assert ("G", "dead") not in output.output[0].columns


def test_crosstab_with_totals_with_suppression_with_mean(data, acro):
    """Test the crosstab with both margins and suppression are true and with aggfunc mean."""
    _ = acro.crosstab(
        data.year,
        data.grant_type,
        values=data.inc_grants,
        aggfunc="mean",
        margins=True,
    )
    output = acro.results.get_index(0)
    assert 8689781 == output.output[0]["All"].iat[0]
    assert 5425170.5 == output.output[0]["All"].iat[6]
    assert "R/G" not in output.output[0].columns


def test_crosstab_with_totals_and_empty_data(data, acro, caplog):
    """Test crosstab when both margins and suppression are true with a disclosive dataset."""
    data = data[
        (data.year == 2010)
        & (data.grant_type == "G")
        & (data.survivor == "Dead in 2015")
    ]
    _ = acro.crosstab(
        data.year,
        [data.grant_type, data.survivor],
        values=data.inc_grants,
        aggfunc="mean",
        margins=True,
    )
    assert (
        "All the cells in this data are disclosive. Thus suppression can not be applied"
        in caplog.text
    )


def test_crosstab_with_manual_totals_with_suppression(data, acro):
    """Test crosstab when margins and suppression are true with the total manual function."""
    _ = acro.crosstab(data.year, data.grant_type, margins=True, show_suppressed=True)
    output = acro.results.get_index(0)
    assert 145 == output.output[0]["All"].iat[0]

    total_rows = output.output[0].iloc[-1, 0:4].sum()
    total_cols = output.output[0].loc[2010:2015, "All"].sum()
    assert 870 == total_cols == total_rows == output.output[0]["All"].iat[6]
    assert "R/G" in output.output[0].columns
    assert np.isnan(output.output[0]["R/G"].iat[0])


def test_crosstab_with_manual_totals_with_suppression_hierarchical(data, acro):
    """Test crosstab when margins and suppression are true with hierarchical data.

    Tests with multilevel indexes and columns while using the total manual function.
    """
    _ = acro.crosstab(
        [data.year, data.survivor],
        [data.grant_type, data.status],
        margins=True,
        show_suppressed=True,
    )
    output = acro.results.get_index(0)
    assert 47 == output.output[0]["All"].iat[0]

    total_rows = (output.output[0].loc[("All", ""), :].sum()) - output.output[0][
        "All"
    ].iat[12]
    total_cols = (output.output[0].loc[:, "All"].sum()) - output.output[0]["All"].iat[
        12
    ]
    assert total_cols == total_rows == output.output[0]["All"].iat[12] == 852
    assert ("G", "dead") in output.output[0].columns
    assert np.isnan(output.output[0][("G", "dead")].iat[0])


def test_crosstab_with_manual_totals_with_suppression_with_aggfunc_mean(data, acro):
    """Test crosstab.

    Tests the crosstab with both margins and suppression are true and with
    aggfunc mean while using the total manual function.
    """
    _ = acro.crosstab(
        data.year,
        data.grant_type,
        values=data.inc_grants,
        aggfunc="mean",
        margins=True,
        show_suppressed=True,
    )
    output = acro.results.get_index(0)
    assert 8689780 == round(output.output[0]["All"].iat[0])
    assert 5425170 == round(output.output[0]["All"].iat[6])
    assert "R/G" in output.output[0].columns
    assert np.isnan(output.output[0]["R/G"].iat[0])


def test_hierarchical_crosstab_with_manual_totals_with_mean(data, acro):
    """Test crosstab.

    Test the crosstab with both margins and suppression are true, with aggfunc
    mean and with multilevel columns and rows while using the total manual
    function.
    """
    _ = acro.crosstab(
        [data.year, data.survivor],
        [data.grant_type, data.survivor],
        values=data.inc_grants,
        aggfunc="mean",
        margins=True,
        show_suppressed=True,
    )
    output = acro.results.get_index(0)
    assert 1385162 == round(output.output[0]["All"].iat[0])
    assert 5434959 == round(output.output[0]["All"].iat[12])
    assert ("G", "Dead in 2015") in output.output[0].columns
    assert np.isnan(output.output[0][("G", "Dead in 2015")].iat[0])


def test_crosstab_with_manual_totals_with_suppression_with_aggfunc_std(
    data, acro, caplog
):
    """Test crosstab.

    Test the crosstab with both margins and suppression are true and with
    aggfunc std while using the total manual function.
    """
    _ = acro.crosstab(
        data.year,
        data.grant_type,
        values=data.inc_grants,
        aggfunc="std",
        margins=True,
        show_suppressed=True,
    )
    output = acro.results.get_index(0)
    assert "All" not in output.output[0].columns
    assert np.isnan(output.output[0]["R/G"].iat[0])
    assert (
        "The margins with the std agg func can not be calculated. "
        "Please set the show_suppressed to false to calculate it." in caplog.text
    )


def test_pivot_table_with_totals_with_suppression(data, acro):
    """Test the pivot table with both margins and suppression are true."""
    _ = acro.pivot_table(
        data,
        index=["year"],
        columns=["grant_type"],
        values=["inc_grants"],
        aggfunc="count",
        margins=True,
    )
    output = acro.results.get_index(0)
    assert 74 == output.output[0][("inc_grants", "All")].iat[0]

    total_rows = output.output[0].iloc[-1, 0:3].sum()
    total_cols = output.output[0].loc[2010:2015, ("inc_grants", "All")].sum()
    assert (
        766
        == total_cols
        == total_rows
        == output.output[0][("inc_grants", "All")].iat[6]
    )
    assert "R/G" not in output.output[0].columns


def test_crosstab_multiple_aggregate_function(data, acro):
    """Crosstab with multiple agg funcs."""
    acro = ACRO(suppress=False)

    _ = acro.crosstab(
        data.year, data.grant_type, values=data.inc_grants, aggfunc=["mean", "std"]
    )
    output = acro.results.get_index(0)
    correct_summary: str = (
        "fail; threshold: 14 cells may need suppressing;"
        " p-ratio: 4 cells may need suppressing; "
        "nk-rule: 2 cells may need suppressing; "
    )
    assert output.summary == correct_summary, (
        f"\n{output.summary}\n should be \n{correct_summary}\n"
    )
    print(f"{output.output[0]['mean']['R/G'].sum()}")
    correctval = 97383496.0
    assert output.output[0]["mean"]["R/G"].sum() == correctval


def test_crosstab_with_totals_with_suppression_with_two_aggfuncs(data, acro):
    """Test crosstab.

    Test the crosstab with both margins and suppression are true and with a
    list of aggfuncs while using the total manual function.
    """
    _ = acro.crosstab(
        data.year,
        data.grant_type,
        values=data.inc_grants,
        aggfunc=["count", "std"],
        margins=True,
    )
    _ = acro.crosstab(
        data.year,
        data.grant_type,
        values=data.inc_grants,
        aggfunc="count",
        margins=True,
    )
    _ = acro.crosstab(
        data.year,
        data.grant_type,
        values=data.inc_grants,
        aggfunc="std",
        margins=True,
    )
    output = acro.results.get_index(0)
    assert output.output[0].shape[1] == 8
    output_1 = acro.results.get_index(1)
    output_2 = acro.results.get_index(2)
    output_3 = pd.concat([output_1.output[0], output_2.output[0]], axis=1)
    output_4 = (output.output[0]).droplevel(0, axis=1)
    assert output_3.equals(output_4)


def test_crosstab_with_totals_with_suppression_with_two_aggfuncs_hierarchical(
    data, acro
):
    """Test crosstab.

    Test the crosstab with both margins and suppression are true and with a
    list of aggfuncs and a list of columns while using the total manual
    function.
    """
    _ = acro.crosstab(
        data.year,
        [data.grant_type, data.survivor],
        values=data.inc_grants,
        aggfunc=["count", "std"],
        margins=True,
    )
    output = acro.results.get_index(0)
    assert ("count", "G", "Alive in 2015") in output.output[0].columns
    assert ("std", "G", "Alive in 2015") in output.output[0].columns


def test_crosstab_with_manual_totals_with_suppression_with_two_aggfunc(
    data, acro, caplog
):
    """Test crosstab.

    Test the crosstab with both margins and suppression are true and with a
    list of aggfuncs while using the total manual function.
    """
    _ = acro.crosstab(
        data.year,
        data.grant_type,
        values=data.inc_grants,
        aggfunc=["count", "std"],
        margins=True,
        show_suppressed=True,
    )
    assert (
        "We can not calculate the margins with a list of aggregation functions. "
        "Please create a table for each aggregation function" in caplog.text
    )


def test_histogram_disclosive(data, acro, caplog):
    """Test a discolsive histogram."""
    filename = os.path.normpath("acro_artifacts/histogram_0.png")
    _ = acro.hist(data, "inc_grants")
    assert os.path.exists(filename)
    acro.add_exception("output_0", "Let me have it")
    results: Records = acro.finalise(path=PATH)
    output_0 = results.get_index(0)
    assert output_0.output == [filename]
    assert (
        "Histogram will not be shown as the inc_grants column is disclosive."
        in caplog.text
    )
    assert output_0.status == "fail"
    shutil.rmtree(PATH)


def test_histogram_non_disclosive(data, acro):
    """Test a non disclosive histogram."""
    filename = os.path.normpath("acro_artifacts/histogram_0.png")
    _ = acro.hist(data, "inc_grants", bins=1)
    assert os.path.exists(filename)
    acro.add_exception("output_0", "Let me have it")
    results: Records = acro.finalise(path=PATH)
    output_0 = results.get_index(0)
    assert output_0.output == [filename]
    assert output_0.status == "review"
    shutil.rmtree(PATH)


def test_finalise_with_existing_path(data, acro, caplog):
    """Test using a path that already exists when finalising."""
    _ = acro.crosstab(data.year, data.grant_type)
    acro.add_exception("output_0", "Let me have it")
    acro.finalise(PATH)
    _ = acro.crosstab(data.status, data.grant_type)
    acro.finalise(PATH)
    assert (
        "Results file can not be created. Directory RES_PYTEST "
        "already exists. Please choose a different directory name." in caplog.text
    )
    shutil.rmtree(PATH)


def test_finalise_non_interactive(data):
    """Test finalise_non_interactive.

    Test that non-interactive version of finalising acro
    leaves exceptions as they were for disclosive table.
    """
    acro = ACRO(suppress=False)
    _ = acro.crosstab(data.year, data.grant_type)
    acro.suppress = True
    _ = acro.crosstab(data.year, data.grant_type)
    # write JSON

    path = "outputs"

    _ = acro.finalise(path, "json", interactive=False)
    result = acro.results

    # load JSON
    loaded: Records = load_records(path)
    orig0 = result.get_index(0)
    read0 = loaded.get_index(0)
    orig1 = result.get_index(1)
    read1 = loaded.get_index(1)
    # check equal
    assert orig0.exception is None or len(orig0.exception) == 0, (
        f"orig exception: expected None, got _{orig0.exception}_"
    )
    assert read0.exception is None or len(read0.exception) == 0, (
        f"read exception: expected None, got _{read0.exception}_"
    )
    assert orig1.exception == "Suppression automatically applied where needed"
    assert read1.exception == "Suppression automatically applied where needed"

    # check SDC outcome DataFrame
    orig_df = orig0.output[0].reset_index()
    read_df = read0.output[0]
    pd.testing.assert_frame_equal(
        orig_df, read_df, check_names=False, check_dtype=False
    )
    if os.path.isdir(path):
        shutil.rmtree(path)


def test_finalise_interactive(data):
    """Test finalise_interactive.

    Test that interactive version of finalising acro
    leaves exceptions as they should be disclosive table.
    """
    acro = ACRO(suppress=False)
    _ = acro.crosstab(data.year, data.grant_type)
    acro.suppress = True
    _ = acro.crosstab(data.year, data.grant_type)
    # write JSON

    mypath = "outputs"

    with patch("builtins.input", return_value="Oh, please..."):
        _ = acro.finalise(mypath, "json", interactive=True)
    result = acro.results
    # load JSON
    loaded: Records = load_records(mypath)
    orig0 = result.get_index(0)
    read0 = loaded.get_index(0)
    orig1 = result.get_index(1)
    read1 = loaded.get_index(1)
    # check equal
    assert orig0.exception == "Oh, please..."
    assert read0.exception == "Oh, please..."
    assert orig1.exception == "Suppression automatically applied where needed"
    assert read1.exception == "Suppression automatically applied where needed"
    print(orig0.exception)
    # check SDC outcome DataFrame
    orig_df = orig0.output[0].reset_index()
    read_df = read0.output[0]
    pd.testing.assert_frame_equal(
        orig_df, read_df, check_names=False, check_dtype=False
    )
    if os.path.isdir(mypath):
        shutil.rmtree(mypath)


def test_create_dataframe(data):
    """Test correct functionality of code to create data frame."""
    # correct
    rows = [data.year, data.grant_type]
    cols = [data.survivor, data.status]
    mydataframe = acro_tables.create_dataframe(rows, cols)
    assert mydataframe.shape == (918, 4)

    # no rows
    mydataframe2 = acro_tables.create_dataframe(None, cols)
    assert list(mydataframe2.columns.values) == ["survivor", "status"]

    # invalid rows
    mydataframe2a = acro_tables.create_dataframe(["year", "grant_type"], cols)
    assert list(mydataframe2a.columns.values) == ["survivor", "status"]

    # no cols
    mydataframe3 = acro_tables.create_dataframe(rows, None)
    assert list(mydataframe3.columns.values) == ["year", "grant_type"]

    # invalid cols
    mydataframe3a = acro_tables.create_dataframe(rows, ["survivor", "status"])
    assert list(mydataframe3a.columns.values) == ["year", "grant_type"]

    # neither
    mydataframe4 = acro_tables.create_dataframe(None, None)
    assert mydataframe4.empty, (
        "dataframe created with no rows or cols should be empty "
        f"but got shape{mydataframe4.shape}"
    )

    # both invalid
    mydataframe4a = acro_tables.create_dataframe(
        ["year", "grant_type"], ["survivor", "status"]
    )
    assert mydataframe4a.empty, (
        "dataframe created with invalid rows  and columns should be empty"
        f"but got shape{mydataframe4a.shape}"
    )


def test_toggle_suppression():
    """Test toggling suppression on/off."""
    acro = ACRO(suppress=False)
    assert not acro.suppress
    acro.enable_suppression()
    assert acro.suppress
    acro.disable_suppression()
    assert not acro.suppress



"""
File with commands to manage the stata-acro interface.

Jim Smith 2023 @james.smith@uwe.ac.uk
MIT licenses apply.
"""

import os
import re

import pandas as pd
import statsmodels.iolib.summary as sm_iolib_summary

from acro import ACRO, acro_regression, add_constant, stata_config
from acro.utils import prettify_table_string


def apply_stata_ifstmt(raw: str, all_data: pd.DataFrame) -> pd.DataFrame:
    """Parse an if statement from stata format then use it to subset a dataframe by contents."""
    if len(raw) == 0:
        return all_data
    # lose any stray 'if' keywords that have got across
    raw = raw.replace("if", "")
    # add braces around each clause- keeping any in the original
    raw = "( " + raw + ")"
    raw = raw.replace("&", ") & (")
    raw = raw.replace("|", ") | (")
    # put spaces around operators to ease parsing
    for operator in [">", "<", "==", ">=", "<=", "!="]:
        raw = raw.replace(operator, " " + operator + " ")

    # apply exclusion
    some_data = all_data.query(raw)
    return some_data


def parse_location_token(token: str, last: int) -> int:
    """
    Parse index position tokens from stata syntax.

    Stata allows f and F for first item  and l/L for last.
    """
    lookup: dict = {"f": 0, "F": 0, "l": last, "L": last}
    if token in ["f", "F", "l", "L"]:
        pos = lookup[token]
    else:
        try:
            pos = int(token)
            if pos > 0:
                pos -= 1
        except ValueError:
            print("valuerror")
            pos = 0
    return pos


def apply_stata_expstmt(raw: str, all_data: pd.DataFrame) -> pd.DataFrame:
    """Parse an in exp statement from stata and use it to subset a dataframe by row indices."""
    last = len(all_data) - 1
    if "/" not in raw:
        pos = parse_location_token(raw, last)
        if pos < 0:
            start = max(0, last + pos + 1)
            end = last
        else:
            start = 0
            end = min(pos, last)

    else:
        token: list = raw.split("/")
        # first index
        start = parse_location_token(token[0], last)
        if start < 0:
            start = last + 1 + start  # -1==last
        # last index
        end = parse_location_token(token[1], last)
        if end < 0:
            end = last + end  # -1==last
        # enforce start <=end
        if start > end:
            end = last

    return all_data.iloc[start : end + 1]


def find_brace_word(word: str, raw: str):
    """Return contents as a list of strings between '(' following a word and the closing ')'.

    First returned value is True/False depending on parsing ok.
    """
    result = []
    idx = raw.find(word)
    if idx == -1:
        return False, f"{word} not found"
    while idx != -1:
        substr = ""
        idx += len(word) + 1
        while idx < len(raw) and raw[idx] != ")":
            substr += raw[idx]
            idx += 1

        if idx == len(raw):
            return False, "phrase not completed"

        result.append(substr)
        idx = raw.find(word, idx)

    return True, result


def extract_aggfun_values_from_options(contents_found, content, varnames) -> dict:
    """
    Extract the variables to aggregate, and the aggregation function.

    Parameters
    ----------
        contents_found:bool
            did the user specify what other put in cells?
        content: list(str)
            what did they ask for?

    Returns
    -------
     cell contents: (dictionary)
    """
    # contents can be variable names or aggregation functions
    cell_content: dict = {"aggfuncs": list([]), "values": list([])}
    if contents_found and len(content) > 0:
        for element in content:
            contents = element.split()
            for word in contents:
                if word in varnames:
                    if word not in cell_content["values"]:
                        cell_content["values"].append(word)
                else:
                    if word not in cell_content["aggfuncs"]:
                        cell_content["aggfuncs"].append(word)
    return cell_content


def parse_table_details(
    varlist: list, varnames: list, options: str, stata_version: float
) -> dict:
    """
    Parse table details from Stata and return as dictionary.

    Calls version relevant to stata version.

    Parameters
    ----------
    varlist : list
        list of variable names the researcher wants to form rows/cols in their table
    varnames : list
        list of variable names present in the data
        i.e. columns in the dataframe
    options : str
        string contains 'options' as parsed by Stata
        (everything after the comma)
        to be interpreted in table design
    stata_version : float
        defines how things should be interpreted
        because Stata changed syntax of the table commandfor version 17 onwards
    """
    details: dict = {"errmsg": ""}

    # get the two lists of what is wanted in the rows and columns
    if stata_version < 17:
        details.update(get_rows_cols_v16(varlist, options))
        contents_found, content = find_brace_word("contents", options)

    else:
        details.update(get_rows_cols_v17on(varlist))
        contents_found, content = find_brace_word("statistic", options)

    # check that these make sense
    if len(details["rowvars"]) == 0 or len(details["colvars"]) == 0:
        details["errmsg"] = (
            "acro does not currently support one dimensional tables. "
            "To calculate cross tabulation, you need to provide at "
            "least one row and one column."
        )
        return details

    allvars = details["rowvars"] + details["colvars"]
    unrecognised_vars = list(set(allvars) - set(varnames))
    if len(unrecognised_vars) > 0:
        details["errmsg"] += (
            f"Error: invalid names in table specifier: {unrecognised_vars}"
        )
        return details

    # get what the user wants displaying in the cells
    details.update(
        extract_aggfun_values_from_options(contents_found, content, varnames)
    )

    # default values
    details["totals"] = False
    details["suppress"] = False
    details["totals"] = "nototals" not in options
    details["suppress"] = "nosuppress" not in options

    return details


def get_rows_cols_v16(varlist: list, options: str) -> dict:
    """Parse stata-16 style table calls.

    Note this is not for latest version of stata, syntax here:
    https://www.stata.com/manuals16/rtable.pdf
    >> table rowvar [colvar [supercolvar] [if] [in] [weight] [, options].
    """
    rows_cols: dict = {"errmsg": "", "rowvars": list([]), "colvars": list([])}

    rows_cols["rowvars"] = [varlist.pop(0)]
    rows_cols["colvars"] = list(reversed(varlist))

    # by() contents are super-rows
    by_found, superrows = find_brace_word("by", options)
    if by_found and len(superrows) > 0:
        for row in superrows:
            extras = row.split()
            for word in extras:
                if word not in rows_cols["rowvars"]:
                    rows_cols["rowvars"].insert(0, word)
    return rows_cols


def stata_details_to_list(mydetails) -> list:
    """Split details which can have different formats."""
    if isinstance(mydetails, str):
        return mydetails.split()
    return list(mydetails)


def get_rows_cols_v17on(varlist: list) -> dict:
    """
    Get table details for the syntax used by stata_version >= 17.

    https://www.stata.com/manuals/tablesintro.pdf
    syntax: table (rowspec) (colspec) [ (tabspec) ] [ if ] [ in ] [ weight ] [, options ].
    """
    rows_cols: dict = {}
    rows_cols["rowvars"] = stata_details_to_list(varlist.pop(0))
    rows_cols["colvars"] = stata_details_to_list(varlist.pop(0))

    if varlist:
        rows_cols["tables"] = varlist.pop(0).split()
        # print(f"table is {details['tables']}")

    return rows_cols


def parse_and_run(  # pylint: disable=too-many-arguments
    mydata: pd.DataFrame,
    command: str,
    varlist_as_str: str,
    exclusion: str,
    exp: str,
    weights: str,
    options: str,
    stata_version: str,
) -> pd.DataFrame:
    """
    Run the appropriate command on a pre-existing ACRO object stata_acro.

    Takes a dataframe and the parsed stata command line.
    Returns the result as a formatted string.
    """
    # sanity checking
    # can only call init if acro object has not been created
    if command != "init" and isinstance(
        stata_config.stata_acro, str
    ):  # pragma: no cover
        return "You must run acro init before any other acro commands"

    # Sometime_TODO de-abbreviate according to
    # https://www.stata.com/manuals13/u11.pdf#u11.1.3ifexp

    fstata_version = float(stata_version)
    varlist: list = varlist_as_str.split()
    # print(varlist)

    # data reduction
    # print(f'before in {mydata.shape}')
    if len(exp) > 0:
        mydata = apply_stata_expstmt(exp, mydata)
    # print(f'after in, before if {mydata.shape}')
    if len(exclusion) > 0:
        mydata = apply_stata_ifstmt(exclusion, mydata)
    # print(f'after both {mydata.shape}')

    # now look at the commands
    session_commands = [
        "init",
        "finalise",
        "enable_suppression",
        "disable_suppression",
        "print_outputs",
    ]
    output_commands = [
        "remove_output",
        "rename_output",
        "add_comments",
        "add_exception",
    ]
    regression_commands = ["regress", "probit", "logit"]
    outcome = ""
    if command in session_commands:
        outcome = run_session_command(command, varlist, options)
    elif command == "custom_output":
        outcome = add_custom_output(varlist)
    elif command in output_commands:
        outcome = run_output_command(command, varlist)
    elif command in regression_commands:
        outcome = run_regression(command, mydata, varlist)
    elif command == "table" and fstata_version == 16:
        outcome = run_table_command(mydata, varlist, weights, options, fstata_version)
    elif command == "table" and fstata_version >= 17:
        varlist = extract_strings(varlist_as_str)
        outcome = run_table_command(mydata, varlist, weights, options, fstata_version)

    else:
        outcome = f"acro command not recognised: {command}"
    return outcome


def run_session_command(command: str, varlist: list, options: str) -> str:
    """Run session commands that are data-independent."""
    outcome = ""

    if command == "init":
        args: dict = {}
        found, config = find_brace_word("config", options)
        if found:
            assert len(config) == 1, "can only supply one config file name"
            args["config"] = config[0]
        found, suppress = find_brace_word("suppress", options)
        if found:
            args["suppress"] = suppress
        stata_config.stata_acro = ACRO(**args)
        outcome = "acro analysis session created\n"

    elif command == "enable_suppression":
        stata_config.stata_acro.suppress = True
        outcome = "suppression toggled on for subsequent commands"
    elif command == "disable_suppression":
        stata_config.stata_acro.suppress = False
        outcome = "suppression toggled off for subsequent commands"
    elif command == "finalise":
        suffix = "json"
        out_dir = "stata_outputs"
        if len(varlist) == 1:
            out_dir = varlist[0]
        if len(varlist) == 2:
            out_dir = varlist[0]
            suffix = varlist[1]
        stata_config.stata_acro.finalise(out_dir, suffix)
        outcome = "outputs and stata_outputs.json written\n"

    elif command == "print_outputs":
        stata_config.stata_acro.print_outputs()

    return outcome


def add_custom_output(varlist: list) -> str:
    """Add a custom output."""
    try:
        the_output = varlist.pop(0)
    except IndexError:
        return "syntax error: please pass the name of the output to be added"

    # .gph extension contain data
    _, file_extension = os.path.splitext(the_output)
    if file_extension == ".gph":
        return "Warning: .gph files may not be exported as they contain data."
    comment_str = " ".join(varlist)
    stata_config.stata_acro.custom_output(the_output, comment_str)
    outcome = f"file {the_output} with comment {comment_str} added to session."
    return outcome


def run_output_command(command: str, varlist: list) -> str:
    """Run outcome-level commands.

    First element of varlist is output affected
    rest (if relevant) is string passed to command.
    """
    try:
        the_output = varlist.pop(0)
    except IndexError:
        return "syntax error: please pass the name of the output to be changed"

    # safety- rest of commands need an existing output to work on
    if (  # pylint:disable=consider-iterating-dictionary
        the_output not in stata_config.stata_acro.results.results.keys()
    ):
        return f"no output with name  {the_output} in current acro session.\n"

    if command == "remove_output":
        stata_config.stata_acro.remove_output(the_output)
        outcome = f"output {the_output} removed.\n"
        return outcome

    # rest of commands need more arguments
    the_str = " ".join(varlist)
    if len(the_str) == 0:
        return f"not enough arguments provided for command {command}.\n"
    if command == "rename_output":
        stata_config.stata_acro.rename_output(the_output, the_str)
        outcome = f"output {the_output} renamed to {the_str}.\n"
    elif command == "add_comments":
        stata_config.stata_acro.add_comments(the_output, the_str)
        outcome = f"Comments added to output {the_output}.\n"
    elif command == "add_exception":
        stata_config.stata_acro.add_exception(the_output, the_str)
        outcome = f"Exception request added to output {the_output}.\n"
    return outcome


def extract_var_within_parentheses(input_string):
    """Extract the words within the first parentheses from a string."""
    string = ""
    string_match = re.match(r"\((.*?)\)", input_string)
    if string_match:
        string = string_match.group(1).strip()
        input_string = input_string[len(string_match.group(0)) :].strip()
    return string, input_string


def extract_var_before_parentheses(input_string):
    """Extract the words before the first parentheses."""
    string = ""
    string_match = re.match(r"^(.*?)\(", input_string)
    if string_match:
        string = string_match.group(1).strip()
        input_string = input_string[len(string_match.group(1)) :].strip()
    return string, input_string


def extract_table_var(input_string):
    """Extract the words within the parentheses.

    If there are no parentheses the string is returned.
    """
    string = ""
    # If the string starts with parentheses
    if input_string.startswith("("):
        string, _ = extract_var_within_parentheses(input_string)
    elif input_string:
        string = input_string.strip()
    return string


def extract_colstring_tablestring(input_string):
    """Extract the column and the tables variables as a string.

    It goes through different options eg. whether the column string is between paranthese or not.
    """
    colstring = ""
    tablestring = ""
    if input_string.startswith("("):
        colstring, input_string = extract_var_within_parentheses(input_string)
        if input_string:
            tablestring = extract_table_var(input_string)

    elif "(" not in input_string:
        words = input_string.split()
        colstring = " ".join(words[:])

    else:
        colstring, input_string = extract_var_before_parentheses(input_string)
        if input_string:
            tablestring = extract_table_var(input_string)
    return colstring, tablestring


def extract_strings(input_string):
    """Extract the index, column and the tables variables as a string.

    It goes through different options eg. whether the index string is between paranthese or not.
    """
    rowstring = ""
    colstring = ""
    tablestring = ""

    # If the string doesn’t have parentheses
    if "(" not in input_string:
        words = input_string.split()
        rowstring = " ".join(words[:-1])
        colstring = words[-1]

    # If the string has parentheses
    else:
        # If there are parentheses at the start of the string
        if input_string.startswith("("):
            rowstring, input_string = extract_var_within_parentheses(input_string)
            colstring, tablestring = extract_colstring_tablestring(input_string)

        else:
            # If there are parentheses at the middle of the string
            rowstring, input_string = extract_var_before_parentheses(input_string)
            colstring, tablestring = extract_colstring_tablestring(input_string)
    varlist = [rowstring, colstring, tablestring]
    return varlist


def creates_datasets(data, details):
    """Return the full dataset if the tables parameter is empty.

    Otherwise, it divides the dataset to small dataset each one is the dataset when
    the tables parameter is equal to one of it is unique values.
    """
    set_of_data = {"Total": data}
    msg = ""
    # if tables var parameter was assigned, each table will
    # be treated as an exclusion which will be applied to the data.
    # The number of datasets will be equal to the number of unique values in the tables var
    # Crosstabulation will be calculate for each dataset
    if "tables" in details and details["tables"] != []:
        # print(f"table is {details['tables']}")
        msg = (
            "You need to manually check all the outputs for the risk of differencing.\n"
        )
        for table in details["tables"]:
            unique_values = data[table].unique()
            # print(f"unique_values are {unique_values}")
            for value in unique_values:
                if isinstance(value, str):
                    exclusion = f"{table}=='{value}'"
                else:  # pragma: no cover
                    exclusion = f"{table}=={value}"
                # print(f"exclusion is {exclusion}")
                my_data = apply_stata_ifstmt(exclusion, data)
                set_of_data[exclusion] = my_data
    return set_of_data, msg


def run_table_command(  # pylint: disable=too-many-locals
    data: pd.DataFrame,
    varlist: list,
    weights: str,
    options: str,
    stata_version: float,
) -> str:
    """Convert a stata table command into an acro.crosstab and return a prettified dataframe."""
    weights_empty = len(weights) == 0
    if not weights_empty:  # pragma
        return f"weights not currently implemented for _{weights}_\n"

    varnames = data.columns
    details = parse_table_details(varlist, varnames, options, stata_version)
    if len(details["errmsg"]) > 0:
        return details["errmsg"]

    aggfuncs = list(map(lambda x: x.replace("sd", "std"), details["aggfuncs"]))
    # don't pass single aggfunc as a list
    if len(aggfuncs) == 1:
        aggfuncs = aggfuncs[0]

    set_of_data, msg = creates_datasets(data, details)

    results = ""
    for exclusion, my_data in set_of_data.items():
        rows, cols = [], []
        # print(f"my data is {my_data}")
        for row in details["rowvars"]:
            rows.append(my_data[row])
        for col in details["colvars"]:
            cols.append(my_data[col])
        # print(f"rows are {rows}")
        # print(f"cols are {cols}")
        if len(aggfuncs) > 0 and len(details["values"]) > 0:
            # sanity checking
            # if len(rows) > 1 or len(cols) > 1:
            #     msg = (
            #         "acro crosstab with an aggregation function "
            #         " does not currently support hierarchies within rows or columns"
            #     )
            #     return msg

            if len(details["values"]) > 1:
                msg = (
                    "pandas crosstab can  aggregate over multiple functions "
                    "but only over one feature/attribute: provided as 'value'"
                )
                return msg
            val = details["values"][0]
            values = data[val]
            print(exclusion)
            safe_output = stata_config.stata_acro.crosstab(
                index=rows,
                columns=cols,
                aggfunc=aggfuncs,
                values=values,
                margins=details["totals"],
                margins_name="Total",
            )

        else:
            print(exclusion)
            safe_output = stata_config.stata_acro.crosstab(
                index=rows,
                columns=cols,
                # suppress=details['suppress'],
                margins=details["totals"],
                margins_name="Total",
            )
        results += f"{exclusion}\n{prettify_table_string(safe_output)}\n"

        options_str = ""
        formatting = [
            "cellwidth",
            "csepwidth",
            "stubwidth",
            "scsepwidth",
            "center",
            "left",
        ]
        if any(word in options for word in formatting):
            options_str = (
                "acro does not currently support table formatting commands.\n "
            )
    return msg + options_str + results


def run_regression(command: str, data: pd.DataFrame, varlist: list) -> str:
    """Interpret and run appropriate regression command."""
    # get components of formula
    depvar = varlist[0]
    indep_vars = varlist[1:]
    new_data = data[varlist].dropna()
    x_var = new_data[indep_vars]
    x_var = add_constant(x_var)
    res_str = ""
    if command == "regress":
        y_var = new_data[depvar]
        results = stata_config.stata_acro.ols(y_var, x_var)
        res_str = get_regr_results(results, "OLS Regression")
    elif command == "probit":
        y_var = new_data[depvar].astype("category").cat.codes  # numeric
        y_var.name = depvar
        results = stata_config.stata_acro.probit(y_var, x_var)
        res_str = get_regr_results(results, "Probit Regression")
    elif command == "logit":
        y_var = new_data[depvar].astype("category").cat.codes  # numeric
        y_var.name = depvar
        results = stata_config.stata_acro.logit(y_var, x_var)
        res_str = get_regr_results(results, "Logit Regression")
    else:  # pragma: no cover
        res_str = f"unrecognised regression command {command}\n"
    return res_str


def get_regr_results(results: sm_iolib_summary.Summary, title: str) -> str:
    """Translate statsmodels.io.summary object into prettified table."""
    res_str = title + "\n"
    for table in acro_regression.get_summary_dataframes(results.summary().tables):
        res_str += prettify_table_string(table, separator=",") + "\n"
        res_str += f"{results.summary().extra_txt}\n"
    return res_str
